from dataclasses import dataclass


@dataclass
class ClientCommand:
    command: str
    args: list


def parse_client_command(message: str) -> ClientCommand:
    """ Will try to match a client command inside a message. A client command begins with a '/' character.
    for example /reset all will be parsed as command='reset' and args={'all'}
    Another example /phone set +1232342342 will be parsed as command='phone' and args={'set', '+1232342342'}
    """
    if message.startswith("/"):
        parts = message.split(" ")
        command = parts[0][1:]
        args = []
        if len(parts) > 1:
            args = parts[1:]
        return ClientCommand(command, args)

    else:
        return None
        


if __name__ == "__main__":
    print(parse_client_command("/reset all"))
    print(parse_client_command("/phone set +1232342342"))





from dataclasses import dataclass
from typing import Any
from cel.assistants.common import FunctionDefinition, Param
from cel.assistants.function_response import FunctionResponse, RequestMode
from cel.assistants.context import Context
from loguru import logger as log

@dataclass
class FunctionContext(Context):
    def __init__(self, *args, params: dict, function_definition: FunctionDefinition, **kwargs):
        super().__init__(*args, **kwargs)
        self.params = params
        self.function_definition = function_definition

    @staticmethod
    def response_text(text: str, request_mode: RequestMode = RequestMode.SINGLE):
        return FunctionResponse(text=text, request_mode=request_mode)

    def validate_params(self, validate_types: bool = False):
        # TODO: Implement validation of params
        assert isinstance(self.params, dict), "Params must be a dictionary"
        assert isinstance(self.function_definition, FunctionDefinition), "Function definition must be an instance of FunctionDefinition"
        
        name = self.function_definition.name
        params_definition = self.function_definition.parameters
        
        if not params_definition or len(params_definition) == 0:
            return
        
        errors = []
        for param in params_definition:
            assert isinstance(param, Param), "Param must be an instance of Param"
            param_name = param.name
            param_type = param.type
            param_value = self.params.get(param_name)
            
            if param_value is None and param.required:
                errors.append(f"Missing required parameter '{param_name}' in function '{name}'")
            
            if param_value is not None:
                if validate_types:        
                    type_error = self.validate_type(param, param_value)
                    if type_error:
                        errors.append(type_error)
                
        return errors
    
    
    def validate_type(self, param: Param, param_value: Any):
        """
            Para types to validate:
                - string: for string values.
                - integer: For int values.
                - number: float for decimal values and int for whole numbers.
                - boolean:for bool values.
                - enum: for enum values.
            
            The value must be converted to the type defined in the function definition, in order to
            test and validate the type.
            
            This function returns an error message if the type is not valid, otherwise it returns None        
        """
        
        try:
            param_type = param.type
            param_name = param.name
            
            if param_type == "string":
                if not isinstance(param_value, str):
                    return f"Parameter '{param_name}' must be a string, but got {type(param_value).__name__}"
                if param.enum and param_value not in param.enum:
                    return f"Invalid value '{param_value}' for parameter '{param_name}'. Expected one of {param.enum}"
                
            elif param_type == "integer":
                if not isinstance(param_value, int):
                    return f"Parameter '{param_name}' must be an integer, but got {type(param_value).__name__}"
            elif param_type == "number":
                if not isinstance(param_value, (int, float)):
                    return f"Parameter '{param_name}' must be a number, but got {type(param_value).__name__}"
            elif param_type == "boolean":
                if not isinstance(param_value, bool):
                    return f"Parameter '{param_name}' must be a boolean, but got {type(param_value).__name__}"
            else:
                log.warning(f"FunctionContext: Type validation not implemented for parameter '{param_name}' with type '{param_type}'")
        except Exception as e:
            return f"Error validating parameter '{param_name}': {str(e)}"
        
        return None

        


import marko
from marko.block import Document
from abc import ABC
from dataclasses import dataclass, field

@dataclass
class Block(ABC):
    type: str
    text: str
    index: int
    breadcrumbs: list[str]

def build_breadcrumbs(doc: Document, current_index: int) -> list[str]:
    breadcrumbs = []
    current_level = float('inf')
    # iterate from current_index to the beginning of the document
    for block in reversed(doc.children[:current_index]):
        if doc.children.index(block) >= current_index:
            break
        if block.get_type().lower() == 'heading':
            level = block.level
            if level < current_level:
                current_level = level
                breadcrumbs.append(block.children[0].children)
    breadcrumbs.reverse()
    return breadcrumbs


def parse_markdown(md: str, split_table_rows: bool = False) -> list[Block]:
    from marko.md_renderer import MarkdownRenderer
    mdr = marko.Markdown(extensions=['gfm'], renderer=MarkdownRenderer)
    doc = mdr.parse(md)
    block_types = ['paragraph', 'code', 'blockquote', 'html', 'hr', 'list', 'listitem', 'table', 'tablerow', 'tablecell', 'strong', 'em', 'codespan', 'br', 'del', 'link', 'image', 'text'];

    blocks = []

    for child in doc.children:

        type = child.get_type().lower()
        
        if (type) in block_types:
            text = child.children[0].children
            child_index = doc.children.index(child)
            bc = build_breadcrumbs(doc, child_index)
            blocks.append(Block(type=type, text=text, index=child_index, breadcrumbs=bc))
            
        if type == 'table':
            table = []
            header = child.children[0].children
            rows = child.children[1:]
            
            if split_table_rows:
                header_text = ' | '.join([cell.children[0].children for cell in header])
                header_sepator = ' | '.join(['---' for cell in header])
                child_index = doc.children.index(child)
                bc = build_breadcrumbs(doc, child_index)            
                
                for row in rows:
                    row_text = ' | '.join([cell.children[0].children for cell in row.children])
                    # print(row_text)
                    table.append(f"{header_text}\n{header_sepator}\n{row_text}")
                    
                for row in table:
                    child_index = doc.children.index(child)
                    bc = build_breadcrumbs(doc, child_index)
                    blocks.append(Block(type='table', text=row, index=child_index, breadcrumbs=bc))
            else:
                table_text = mdr.render(child)
                bc = build_breadcrumbs(doc, child_index)
                blocks.append(Block(type='table', text=table_text, index=doc.children.index(child), breadcrumbs=bc))

                

    return blocks


#   {
#     "type": "document",
#     "document": {
#       "link": "the-provider-name/protocol://the-url",
#       # provider and filename are optional parameters
#       "provider": {
#         "name" : "provider-name"
#       },
#       "filename": "your-document-filename"
#     }
#   }
from .component import Component
from .utils import del_none
import json



class DocumentURL(Component):
    def __init__(self, link: str, provider: str | None = None, filename: str | None = None) -> None:
        self.link = link
        self.provider = provider
        self.filename = filename

    def __str__(self):
        document = {
            "type": "document",
            "document": {
                "link": self.link,
                "provider": {"name": self.provider} if self.provider else None,
                "filename": self.filename
            }
        }
        del_none(document)
        return json.dumps(document)

    def __repr__(self):
        return f"DocumentURLComponent(link={self.link}, provider={self.provider}, filename={self.filename})"

    def __eq__(self, other):
        return self.link == other.link and self.provider == other.provider and self.filename == other.filename


import json
import secrets
import time
import jwt
import datetime
from cryptography.fernet import Fernet
import base64


def generate_encryption_key():
    return secrets.token_bytes(32)

def generate_secret_key():
    return secrets.token_hex(32)



def create_jwt(data: dict, secret_key, encryption_key, ttl=180):
    """
    Creates an encrypted JWT (JSON Web Token).

    Args:
        data (dict): Data to include in the JWT.
        secret_key (str): Secret key to sign the JWT.
        encryption_key (str): Key to encrypt the JWT.
        ttl (int, optional): Token time-to-live in seconds. Default is 180 seconds.

    Returns:
        str: Encrypted JWT as a string.
    """    
    # Calculate the expiration time
    expiration = datetime.datetime.utcnow() + datetime.timedelta(seconds=ttl)

    try:
        payload = data or {}
    except ValueError as e:
        raise ValueError("El string proporcionado no es un JSON válido.") from e

    # Add expiration time to the payload
    payload['exp'] = expiration

    # Sign the JWT with the secret key using the HS256 algorithm
    encoded_jwt = jwt.encode(payload, secret_key, algorithm='HS256')

    # Encrypt the JWT with the encryption key
    fernet = Fernet(base64.urlsafe_b64encode(encryption_key.ljust(32, b'\0')))
    encrypted_jwt = fernet.encrypt(encoded_jwt.encode())

    # Return the encrypted JWT as a string
    return encrypted_jwt.decode()


def decode_jwt(jwt_token, secret_key, encryption_key):
    """
    Decodes a JWT token using the provided secret key and encryption key.

    Args:
        jwt_token (str): The JWT token to decode.
        secret_key (str): The secret key used to verify the JWT signature.
        encryption_key (bytes): The encryption key used to decrypt the JWT.

    Returns:
        dict: The decoded JWT payload.

    Raises:
        ValueError: If the token has expired or is invalid.
    """

    # Unencrypt the JWT
    fernet = Fernet(base64.urlsafe_b64encode(encryption_key.ljust(32, b'\0')))
    decrypted_jwt = fernet.decrypt(jwt_token.encode())

    # Verify the JWT signature, exp date and decode it
    try:
        decoded_jwt = jwt.decode(decrypted_jwt, secret_key, algorithms=['HS256'])
        return decoded_jwt
    except jwt.ExpiredSignatureError:
        raise ValueError("El token ha expirado.")
    except jwt.InvalidTokenError as e:
        raise ValueError("El token no es válido.") from e
    

def generate_link(base_url: str, secret_key: str, encryption_key: bytes, data: dict = None, ttl=180):
    """
    Generates a URL with a JWT token as a query parameter.

    Args:
        base_url (str): The base URL to which the JWT token will be appended.
        secret_key (str): The secret key used to sign the JWT.
        encryption_key (bytes): The encryption key used to encrypt the JWT.
        data (dict, optional): The data to include in the JWT payload. Defaults to None.
        ttl (int, optional): The time-to-live for the JWT token in seconds. Defaults to 180.

    Returns:
        str: The generated URL with the JWT token as a query parameter.
    """    
    
    assert isinstance(base_url, str), "base_url must be a string"
    assert isinstance(secret_key, str), "secret_key must be a string"
    assert isinstance(encryption_key, bytes), "encryption_key must be a bytes object"
    assert isinstance(data, dict) or None, "data must be a dictionary"
    

    # Create the JWT token
    jwt_token = create_jwt(data, secret_key, encryption_key, ttl)

    # Return the URL with the JWT token as a query parameter
    return f"{base_url}/{jwt_token}"



if __name__ == "__main__":
    
    # Use example
    secret_key = generate_secret_key()
    encryption_key = generate_encryption_key()
    
    
    data = {"user_id": "123", "name": "John Doe"}

    jwt_token = create_jwt(data, secret_key, encryption_key, ttl=1)
    print(jwt_token)

    time.sleep(2)

    # Decodificar el JWT
    try:
        decoded_token = decode_jwt(jwt_token, secret_key, encryption_key)
        print(decoded_token)
    except ValueError as e:
        print(f"Error: {e}")
    

    # Generate a link
    base_url = "https://example.com"
    data = {"user_id": "123", "name": "John Doe"}
    link = generate_link(base_url, secret_key, encryption_key, data)
    print(link)
    




def filter_phone_number(phone_number: str):
    if phone_number is None:
        return None
    
    if phone_number.startswith('54911'):
        # replace 54911 with 5411
        return '5411' + phone_number[5:]
    
    if phone_number.startswith('521'):        
        return phone_number.replace('521', '52')
    
    return phone_number




from typing import List, Optional
from dataclasses import dataclass, field
from cel.model.common import ContextMessage
from cel.rag.providers.rag_retriever import RAGRetriever
from cel.rag.stores.vector_store import VectorRegister
from langchain_openai import ChatOpenAI
from langchain_core.messages import (
    SystemMessage
)

DEFAULT_PROMPT = (
    "You are an AI assistant specialized in transforming conversation inputs into well-structured and context-aware queries for retrieval-augmented generation (RAG). "
    "Your goal is to analyze the conversation history and refine the user's query to maximize its relevance and effectiveness for vectorization and retrieval, without changing the original intent. "
    "Focus on capturing the key context, entities, and intent from the conversation to generate concise, precise, and highly actionable queries. "
    "Avoid introducing unnecessary complexity or irrelevant details. When refining, ensure the enhanced query naturally aligns with the context and flow of the conversation."
)

@dataclass
class QueryRefiner:
    model_name: str = "gpt-3.5-turbo"
    n_history_messages: int = 3
    custom_prompt: str = DEFAULT_PROMPT
    temperature: float = 0.0
    max_tokens: int = 50
    llm: Optional[ChatOpenAI] = field(init=False, default=None)

    def __post_init__(self):
        self.llm = ChatOpenAI(
            model=self.model_name,
            temperature=self.temperature,
            max_tokens=self.max_tokens
        )

    def enhance_query(self, query: str, history: List[ContextMessage]) -> str:

        enhanced_query = query

        if history:
            history_context = "\n".join(
                [f"User: {msg.content}" for msg in history[-self.n_history_messages:]])
            prompt = f"{self.custom_prompt}\n\nHistory:\n{history_context}\n\nUser Query: {query}\n\nEnhanced Query:"

            response = self.llm.invoke([SystemMessage(content=prompt)])

            if response and isinstance(response, SystemMessage):
                enhanced_query = response.content

        return enhanced_query


class EnhancedRetriever(RAGRetriever):
    def __init__(self, base_retriever: RAGRetriever, query_builder: QueryRefiner = None):
        self.base_retriever = base_retriever
        self.query_builder = query_builder or QueryRefiner()

    def search(self,
               query: str,
               top_k: int = 1,
               history: List[ContextMessage] = None,
               state: dict = {}) -> List[VectorRegister]:
        # Use the QueryBuilder to enhance the query
        enhanced_query = self.query_builder.enhance_query(query, history)
        # Call the base retriever with the enhanced query
        return self.base_retriever.search(enhanced_query, top_k)



import asyncio
from dataclasses import dataclass
import json
import traceback
from loguru import logger as log
from cel.assistants.function_response import FunctionResponse
from cel.assistants.common import EventResponse
from cel.assistants.macaw.custom_chat_models.chat_open_router import ChatOpenRouter
from cel.assistants.macaw.macaw_inference_context import MacawNlpInferenceContext
from cel.assistants.macaw.macaw_history_adapter import MacawHistoryAdapter
from cel.assistants.macaw.macaw_utils import get_last_n_elements, map_functions_to_tool_messages
from cel.assistants.stream_content_chunk import StreamContentChunk
from langchain_openai import ChatOpenAI
from langchain_core.messages import HumanMessage, ToolMessage, AIMessageChunk
from langchain_core.messages import (
    SystemMessage,
    message_to_dict,
    messages_from_dict,
)
from langsmith import traceable

DEFAULT_PROMPT = "Create an assistant funny and sarcastic, dark humor chatbot."
LLM_DEFAULT_KWARGS = {
    "model": "gpt-4o",
    "temperature": 0,
    "max_tokens": None,
    "timeout": 20,
    "max_retries": 3,
    "streaming": True,
    "verbose": True
}

@dataclass
class MacawFunctionCall:
    name: str
    args: dict
    id: str     

@traceable
async def process_new_message(ctx: MacawNlpInferenceContext, message: str, on_function_call=None):
    assert isinstance(ctx, MacawNlpInferenceContext),\
        "ctx must be an instance of MacawNlpInferenceContext"
    assert isinstance(message, str),\
        "message must be a string"

    
    history_store = MacawHistoryAdapter(ctx.history_store)
    
    # Create settings
    settings = {**LLM_DEFAULT_KWARGS}
    settings["model"] = ctx.settings.core_model
    settings["temperature"] = ctx.settings.core_temperature
    settings["max_tokens"] = ctx.settings.core_max_tokens
    settings["timeout"] = ctx.settings.core_timeout
    settings["max_retries"] = ctx.settings.core_max_retries
    # **{"model": "mistralai/mixtral-8x7b-instruct"}
    # merge kwargs
    ctx.llm = ctx.llm or ChatOpenAI
    llm = ctx.llm(
        **{**settings, **(ctx.llm_kwargs or {})}
    )

    try:
        # Toolling
        functions = ctx.functions
        if functions is not None and len(functions) > 0:
            mapfuncs = map_functions_to_tool_messages(functions)
            llm_with_tools = llm.bind_tools(mapfuncs)
        else:
            llm_with_tools = llm
    except Exception as e:
        if isinstance(e, NotImplementedError):
            log.error(f"Error binding tools: Functions not implemented")
        else:
            log.error(f"Error binding tools: {e}")
        llm_with_tools = llm    
    
    # Build State
    # ------------------------------------------------------------------------
    try:
        stored_state = await ctx.state_store.get_store(ctx.lead.get_session_id()) or {}    
        # TODO: Remove this when the self.init_state is removed
        stored_state.update(ctx.init_state or {})
    except Exception as e:
        raise ValueError("Macaw NLP: Error getting stored state") from e



    # Compile prompt
    # ------------------------------------------------------------------------  
    try: 
        prompt = await ctx.prompt.compile(stored_state, ctx.lead, message=message)
    except Exception as e:
        raise ValueError("Macaw NLP: Error compiling prompt") from e
    
    # RAG
    # ------------------------------------------------------------------------
    
    # Prompt > System Message
    history = [SystemMessage(prompt)]

    # Load messages from store before RAG retrieval
    msgs = await history_store.get_history(ctx.lead) or []

    # append to messages
    history.extend(msgs)

    # Add the human message
    input_msg = HumanMessage(message)
    
    # New messages buffer is a list of messages to be added to the history
    # If everything goes well, we will append this list to the history
    # at the end of the process
    new_messages = [input_msg]
    
    # Slice the messages 
    try:
        history = get_last_n_elements(history, ctx.settings.core_history_window_length)
    except Exception as e:
        log.critical(f"Error get_last_n_elements: {e}")
        # TODO: if we fail to get the last n elements, we should not process the message?
        # We should return an error message to the user?
        # or keep on processing the whole history?
        # For now, we keep on processing the whole history

    if ctx.rag_retriever:
        rag_response = ctx.rag_retriever.search(message, ctx.settings.core_rag_knn, history + new_messages)
        if rag_response:
            for vr in rag_response:
                prompt += f"\n{vr.text or ''}" 

    response = None
    try:
        # Process LLM invoke in a stream
        async for delta in llm_with_tools.astream(history + new_messages):
            assert isinstance(delta, AIMessageChunk)
            if response is None:
                response = delta
            else:
                response += delta
                
            # Yield the response if it's not a tool call
            if not response.tool_calls:
                yield StreamContentChunk(content=delta.content, is_partial=True)
                
                
        # Now we have the full response
        # Append Response to history (only memory)
        new_messages.append(response)
        
        # if it's a tool call
        if response.tool_calls:
            # We have tool calls, we need to process them
            # Handle multiple function calls and chained calls 
            # We use a for... to handle chained function calls until 
            # the max number of calls is reached
            # easy way to avoid infinite loop$
                
            cancel_ai = False
            for idx in range(ctx.settings.core_max_function_calls_in_message):
                if response.tool_calls:
                    # Do all function calls
                    for tool_call in response.tool_calls: 
                        name = tool_call.get("name")
                        args = tool_call.get("args")
                        id = tool_call.get("id")
                        log.debug(f"Function: {name} called with params: {args}")
                        try:
                            mtool_call = MacawFunctionCall(name, args, id)
                            func_output = await on_function_call(ctx, mtool_call)
                            
                            response_text = None
                            if isinstance(func_output, FunctionResponse):
                                response_text = func_output.text
                            elif isinstance(func_output, EventResponse):
                                if func_output.disable_ai_response:
                                    cancel_ai = True
                                    log.debug(f"Function {name} requested cancel_ai_response")
                                response_text = "ok"
                            elif isinstance(func_output, str):
                                response_text = func_output
                            else:
                                response_text = "Data not found"

                            msg = ToolMessage(response_text, tool_call_id=id)
                            new_messages.append(msg)
                            log.debug(f"History udpated: func: {name} called with params: {args} -> {response_text}")

                        except Exception as e:
                            log.critical(f"Error calling function: {name} with args: {args} - {e}")
                            tool_output = "In this moment I can't process this request."
                            msg = ToolMessage(tool_output, tool_call_id=id)
                            new_messages.append(msg)

                            # break
                            # NOTE: If one function fails, the rest of the functions are not called?
                            # This is a design decision, we can change it later.
                            # ----------
                            # NOTE: Removed due broken behavior. If one function fails, the history ends badly. 
                            # ToolCall messages end with no ToolMessage. It's better to have a message with the error.

                    # Si cancel_ai_response fue solicitado, no re-invocar al LLM
                    if cancel_ai:
                        log.debug("cancel_ai_response: skipping further LLM invocation")
                        break

                    # Process response
                    response = llm_with_tools.invoke(history + new_messages)
                    new_messages.append(response)
                else:
                    yield StreamContentChunk(content=response.content, is_partial=True)
                    break
                
            # TODO: This validation may not be needed
            # -----------------------------------------------
            # # Here we have finished the loop and we got the whole list of new_messages to be stored in the history
            # # Last step is to verify that the new_messages list is valid and store it in the history
            # # Validating the list is important to avoid storing invalid messages sequences in the history
            # # Corrupted history can lead to a blocking error for this user in the future
            # # First message must be a HumanMessage
            # if new_messages[0].type != "human":
            #     raise ValueError("Macaw NLP process_message: First message must be a HumanMessage")
            
            # # Second message must be a AIMessageChunk
            # # with additional arguments for tool calls
            # if not isinstance(new_messages[1], AIMessageChunk):
            #     raise ValueError("Macaw NLP process_message: Second message must be a AIMessageChunk")
            # else:
            #     if not new_messages[1].tool_calls:
            #         raise ValueError("Macaw NLP process_message: Second message must have tool calls")
                
            # tool_calls = new_messages[1].tool_calls
            
            # # Each message from 2 to n-1 must be a ToolMessage
            # for msg in new_messages[2:-1]:
            #     if not isinstance(msg, ToolMessage):
            #         raise ValueError("Macaw NLP process_message: Messages from 2 to n-1 must be ToolMessages")
            
            # # The number of tool calls must be the same as the number of ToolMessages
            # if len(tool_calls) != (len(new_messages)-2):
            #     raise ValueError("Macaw NLP process_message: Number of tool calls must be the same as the number of ToolMessages") 

            
            # TODO: add support for storing batch messages
            for msg in new_messages:
                await history_store.append_to_history(ctx.lead, msg)

            log.debug(f"Validated history store udpated with tool calls: {len(new_messages)} messages stored, session: {ctx.lead.get_session_id()}")
        else:
            # No tool calls, we can store the new_messages in the history
            # TODO: add support for storing batch messages
            for msg in new_messages:
                await history_store.append_to_history(ctx.lead, msg)
            log.debug(f"History store udpated: {len(new_messages)} messages stored, session: {ctx.lead.get_session_id()}")

            
    except Exception as e:
        # Leave the unhandled user message in the history?????
        # TODO: Check if this is the correct behavior
        await history_store.append_to_history(ctx.lead, response)
        
        raise ValueError("Macaw NLP: Error processing message") from e

        
    


@traceable
async def blend_message(ctx: MacawNlpInferenceContext, message: str, history_length: int = 10):
    """Blend a message using the conversation context."""
    
    assert isinstance(ctx, MacawNlpInferenceContext),\
        "ctx must be an instance of MacawNlpInferenceContext"
    assert isinstance(message, str),\
        "message must be a string"
        
    history_store = MacawHistoryAdapter(ctx.history_store)
    
    # Create settings
    settings = {**LLM_DEFAULT_KWARGS}
    settings["model"] = ctx.settings.blend_model
    settings["temperature"] = 0.0
    settings["max_tokens"] = ctx.settings.blend_max_tokens
    settings["timeout"] = ctx.settings.blend_timeout
    settings["max_retries"] = ctx.settings.blend_max_retries
    
    # merge kwargs
    ctx.llm = ctx.llm or ChatOpenAI
    llm = ctx.llm(
        **{**settings, **(ctx.llm_kwargs or {})}
    )


    # Load messages from store
    msgs = await history_store.get_last_messages(
        ctx.lead,
        history_length or ctx.settings.blend_history_window_length) or []
    
    # Map to BaseMessages and append to messages
    dialog = ""
    for msg in msgs:
        dialog += f"{msg.type}: {msg.content}\n"


    system_message = ctx.settings.blend_prompt

    # Dynamically generated with current context and message:
    prompt_message = (
        f"Conversation context:\n{dialog}\n\n"
        f"System message to adapt:\n\"{message}\""
    )

    # Using standard langchain or similar LLM message format:
    messages = [
        SystemMessage(system_message),
        HumanMessage(prompt_message)
    ]

    res = llm.invoke(messages)    

    return res.content

@traceable
async def process_insights(ctx: MacawNlpInferenceContext, targets: dict = {}, history_length: int = 10):
    """ Get insights from the conversation history """
    
    assert isinstance(ctx, MacawNlpInferenceContext),\
        "ctx must be an instance of MacawNlpInferenceContext"
    assert isinstance(targets, dict),\
        "targets must be a dictionary"

    if not targets:
        return None
        
    history_store = MacawHistoryAdapter(ctx.history_store)
    
    # Create settings
    settings = {**LLM_DEFAULT_KWARGS}
    settings["model"] = ctx.settings.insights_model
    settings["temperature"] = 0.0
    settings["max_tokens"] = ctx.settings.insights_max_tokens
    settings["timeout"] = ctx.settings.insights_timeout
    settings["max_retries"] = ctx.settings.insights_max_retries
    
    # merge kwargs
    llm = ChatOpenAI(
        **settings
    )


    # Load messages from store
    msgs = await history_store.get_last_messages(
        ctx.lead, 
        ctx.settings.insights_history_window_length) or []
    # Map to BaseMessages and append to messages
    
    dialog = ""
    for msg in msgs:
        dialog += f"{msg.type}: {msg.content}\n"

    entities = ""
    for target in targets:
        entities += f"{target}: {targets.get(target)}\n"
        
    prompt = f"""Given this conversation: 
{dialog}
Extract insights for the following entities:
{entities}
Build a json (don't use markdown), skip keys with empty values, use the following format:
{{"entity1": "value1", "entity2": "value2"}}"""

    # Prompt > System Message
    messages = [SystemMessage(prompt)]

    try:
        res = await llm.ainvoke(messages)
        return json.loads(res.content)    
    except Exception as e:
        log.error(f"Error processing insights: {e}")
        return None


# DEPERECATED SYNC VERSION

# import json
# from redis import Redis
# from loguru import logger as log
# from prompter.stores.history.base_history_provider import BaseHistoryProvider


# class RedisHistoryProvider(BaseHistoryProvider):

#     def __init__(self, redis: str | Redis, key_prefix: str = "h"):
#         self.client = redis if isinstance(redis, Redis) else Redis.from_url(redis)
#         self.key_prefix = key_prefix
#         log.debug(f"Create: RedisHistoryProvider")

#     def get_key(self, sessionId: str):
#         return f"{self.key_prefix}:{sessionId}"

#     def append_to_history(self, sessionId: str, entry, metadata=None, ttl=None):
#         key = self.get_key(sessionId)
#         value = json.dumps(entry)
#         self.client.rpush(key, value)

#         # set expiration to 24 hours
#         self.client.expire(key, ttl if ttl else 86400)

#     def get_history(self, sessionId: str):
#         key = self.get_key(sessionId)
#         values = self.client.lrange(key, 0, -1)
#         res = [json.loads(v) for v in values]
#         # remove None elements
#         res = [r for r in res if r]

#         return res
        

#     def clear_history(self, sessionId: str, keep_last_messages=None):
#         key = self.get_key(sessionId)
#         if keep_last_messages:
#             self.client.ltrim(key, 0, keep_last_messages)
#         else:
#             self.client.delete(key)

#     def get_history_slice(self, sessionId: str, start, end):
#         key = self.get_key(sessionId)
#         history = self.client.lrange(key, start, end)
#         return [json.loads(h) for h in history]

#     def get_last_messages(self, sessionId: str, count):
#         key = self.get_key(sessionId)
#         history = self.client.lrange(key, -count, -1)
#         return [json.loads(h) for h in history]

#     def close_conversation(self, sessionId: str):
#         raise NotImplementedError("Method not implemented.")
    




import inspect
import json
from loguru import logger as log

from langchain.load.load import load
from langchain.load.dump import dumps

from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message
from cel.stores.history.base_history_provider import BaseHistoryProvider
from cel.stores.state.base_state_provider import BaseChatStateProvider
from cel.stores.history.history_inmemory_provider import InMemoryHistoryProvider
from cel.stores.state.state_inmemory_provider import InMemoryStateProvider
from cel.assistants.macaw.macaw_history_adapter import MacawHistoryAdapter


class LogicRouter(BaseAssistant):

        
    def __init__(self, 
                 assistants: list[BaseAssistant], 
                 assistant_selector_func: callable,
                 history_store: BaseHistoryProvider = None,
                 state_store: BaseChatStateProvider = None
                ):
        
        
        super().__init__(name="Logic Router Assistant", description="This assistant routes the user to the most suitable assistant based on a logic function")
        log.debug(f"Logic Router Assistant created with {len(assistants)} assistants")
        assert len(assistants) > 0, "At least one assistant is required"
        assert callable(assistant_selector_func), "Assistant selector function must be a callable"
        
        # List of assistants
        self._assistants = assistants
        self._assistant_selector_func = assistant_selector_func
        
        # Init state and history store
        self._state_store = state_store or InMemoryStateProvider()
        self._history_store = history_store or InMemoryHistoryProvider()      
        
        # Make sure than all assistants share the same state and history store
        for ast in self._assistants:
            log.debug(f"Assistant Name: {ast.name} -> {ast.description}")
            
            if ast._state_store != self._state_store:
                # TODO: evaluate if this is the correct behavior
                # If the assistant has a different state store, overwrite it???
                log.critical(f"Assistant {ast.name} has different state store")
                ast.set_state_store(self._state_store)
                
            if ast._history_store != self._history_store:
                # If the assistant has a different history store, overwrite it
                # is important to keep the history store consistent
                # History consistency is important for the router to work properly
                log.warning(f"Assistant {ast.name} has different history store")
                ast.set_history_store(self._history_store)
                log.warning(f"Assistant {ast.name} history store overwritten!!")
            
            
    async def __call_function(self, 
                              lead: ConversationLead = None,
                              message: str = None) -> BaseAssistant:
        
        # Get the current state for this lead
        current_state = await self._state_store.get_store(lead.get_session_id()) or {}
        
        args_dict = {
            'lead': lead,
            'session_id': lead.get_session_id(),
            'message': message,
            'state': current_state,
            'assistants': self._assistants
        }

        func = self._assistant_selector_func

        args = inspect.getfullargspec(func).args
        # Build kwargs using args and args_dict
        kwargs = {}
        for arg in args:
            if arg in args_dict:
                kwargs[arg] = args_dict[arg]

        # Check if the function is a coroutine (async function)
        response = None
        if inspect.iscoroutinefunction(func):
            response = await func(**kwargs)
        else:
            response = func(**kwargs)

        assert isinstance(response, BaseAssistant), "Assistant selector function must return a BaseAssistant instance"
        return response

            
            
    async def get_assistant(self, lead: ConversationLead, message: str):
        return await self.__call_function(lead, message)
        

    async def new_message(self, message: Message, local_state: dict = {}):
        log.debug(f"Router Assistant: new message: {message}")
        lead = message.lead
        # Get the assistant
        ast = await self.get_assistant(lead, message)
        assert isinstance(ast, BaseAssistant), "Agent must be a BaseAssistant instance"
        log.debug(f"Router Assistant selected: {ast.name}")
        
        # Call 'message' event in the selected assistant
        res = await ast.call_event("message", lead, message)
        # assert res is None or isinstance(res, EventResponse)
        
        if res is None:
            log.debug(f"Event: 'message' response is None for message: {message.text}")
            
        # if res is not noe and res has attr disable_ai_response, and  res.disable_ai_response is True
        if res and hasattr(res, "disable_ai_response") and res.disable_ai_response:
            log.warning(f"Event 'message' response has disabled AI response for message: {message.text}")
            # Return empty response
            return
            
        # Call the assistant new_message method
        async for chunk in ast.new_message(message, local_state):
            yield chunk
            

                        
        
       
    async def blend(self, lead: ConversationLead, text: str, history_length: int = None):
        log.debug(f"Router Assistant: blend: {text}")
        ast = await self.get_assistant(lead, text)
        
        return await ast.blend(lead, text, history_length)

    
    async def do_insights(self, lead: ConversationLead, targets: dict = {}, history_length: int = 10):
        log.warning("Router Assistant: do insights not implemented")
        return {}
    

    async def append_message_to_history(self, lead: ConversationLead, message: str, role: str = "assistant"):
        assert role in ["assistant", "user", "system"], "role must be one of: assistant, user, system"

        history = MacawHistoryAdapter(self._history_store)
        
        from langchain_core.messages import HumanMessage, SystemMessage, AIMessage
        entry = None
        if role == "assistant":
            entry = AIMessage(message)
        elif role == "user":
            entry = HumanMessage(message)
        elif role == "system":
            entry = SystemMessage(message)
            
        await history.append_to_history(lead, entry)
        
        
    async def process_client_command(self, lead: ConversationLead, command: str, args: list[str]):
    
        if command == "reset":
            
            if args and args[0] == "all":
                await self._history_store.clear_history(lead.get_session_id())
                yield "History cleared"
                await self._state_store.clear_store(lead.get_session_id())
                yield "State cleared"
                return

            await self._history_store.clear_history(lead.get_session_id())
            yield "History cleared"
            return
        
        if command == "state":
            state = await self._state_store.get_store(lead.get_session_id())
            if state is None:
                yield "No state found"
                return
            for k, v in state.items():
                yield f"{k}: {v}"
            return
        
        if command == "history":
            history = await self._history_store.get_history(lead.get_session_id()) or []    

            if history is None or len(history) == 0:
                yield "History is empty"
                return

            for h in history:
                aux = load(h)
                log.debug(f"History: {aux}")
                yield dumps(aux)
            
            return
        
        if command == "set":
            if len(args) < 2:
                yield "Not enough arguments"
                return
            key = args[0]
            value = args[1]
            state = await self._state_store.get_store(lead.get_session_id())
            state[key] = value
            await self._state_store.set_store(lead.get_session_id(), state)
            yield f"State updated: {key}: {value}"
            return
        
        if command == "prompt":
            yield "Macaw Assistant v0.1"

            if args and args[0] == "all":            
                prompt = self.prompt
                # split into 250 chars aprox chunks
                for i in range(0, len(prompt), 250):
                    yield prompt[i:i+250]
                return
            
            # first 250 chars
            yield self.prompt[:250]
            return

            
            


from abc import ABC
import datetime
from cel.gateway.model.attachment import FileAttachment
from cel.gateway.model.conversation_lead import ConversationLead

class OutgoingMessageType:
    TEXT = "text"
    IMAGE = "image"
    AUDIO = "audio"
    VOICE = "voice"
    VIDEO = "video"
    DOCUMENT = "document"
    LOCATION = "location"
    CONTACT = "contact"
    
    SELECT = "select"
    LINK = "link"
    BUTTONS = "buttons"


class OutgoingMessage(ABC):
    """This class represents a generic outgoing message object"""
    
    def __init__(self, 
                #  outgoing message type:
                 type: str,
                 lead: ConversationLead, 
                 metadata: dict = None, 
                 attachments: list[FileAttachment] = None,
                 is_partial: bool = True,
                 is_private: bool = False
                ):
        self.date = datetime.datetime.now().timestamp()
        self.attachments: list[FileAttachment] = attachments
        self.lead = lead
        self.metadata = metadata
        self.is_partial = is_partial
        self.is_private = is_private
        self.type = type
        
        assert isinstance(self.lead, ConversationLead),\
            "lead must be an instance of ConversationLead"
        assert self.metadata is None or isinstance(self.metadata, dict),\
            "metadata must be a dictionary"
        assert self.attachments is None or isinstance(self.attachments, list),\
            "attachments must be a list"
        assert self.type in [OutgoingMessageType.TEXT, 
                        OutgoingMessageType.IMAGE, 
                        OutgoingMessageType.AUDIO, 
                        OutgoingMessageType.VIDEO, 
                        OutgoingMessageType.DOCUMENT, 
                        OutgoingMessageType.LOCATION, 
                        OutgoingMessageType.CONTACT, 
                        OutgoingMessageType.SELECT, 
                        OutgoingMessageType.BUTTONS,
                        OutgoingMessageType.LINK],\
            "type must be a valid OutgoingMessageType"
        
        
        
        


from .model.media_utils import upload_media, delete_media, query_media_url, download_media
from .model.whatsapp_attachment import WhatsappAttachment
from .model.whatsapp_lead import WhatsappLead
from .model.whatsapp_message import WhatsappMessage
from .whatsapp_connector import WhatsappConnector



from dataclasses import dataclass


@dataclass
class MacawSettings:
    """ Macaw Assistant Settings 
    Core settings are used to configure the core new message processing. 
    Blend settings are used to configure the blend message processing.
    Insights settings are used to configure the insights processing.
    """
    core_model: str = "gpt-4o"
    """The model to use for the core processing."""
    core_temperature: float = 0
    """The temperature to use for the core processing"""
    core_max_tokens: int = 200
    """The max tokens to use for the core processing."""
    core_history_window_length: int = 10
    """The history window length to use for the core processing."""
    core_max_function_calls_in_message: int = 5
    """The max number of function calls recursively allowed in a single message."""
    core_max_retries: int = 3
    """The max number of retries allowed for a single message."""
    core_timeout: int = 20
    """The timeout to use for the core processing."""
    core_rag_knn: int = 3
    """The number of nearest neighbors to use for the RAG retrieval."""

    blend_model: str = "gpt-3.5-turbo"
    """The temperature to use for the blend processing"""
    blend_max_tokens: int = 200
    """The max tokens to use for the blend processing."""
    blend_history_window_length: int = 3
    """The history window length to use for the blend processing."""
    blend_timeout: int = 20
    """The timeout to use for the blend processing."""
    blend_max_retries: int = 3
    """Blend custom prompt"""
    blend_prompt: str = (
        "You are responsible for clearly rewriting and adapting technical system messages "
        "into natural, user-friendly text. Translate the messages according to the user's language "
        "and conversational tone. Do not include markdown, technical labels (e.g. user:, assistant:). "
        "Adapt the wording and style to match naturally with conversational context."
    )

    insights_enabled: bool = True
    """Whether to enable the insights processing."""
    insights_model: str = "gpt-3.5-turbo"
    """The temperature to use for the insights processing"""
    insights_max_tokens: int = 300
    """The max tokens to use for the insights processing."""
    insights_history_window_length: int = 3
    """The history window length to use for the insights processing."""
    insights_timeout: int = 20  
    """The timeout to use for the insights processing."""
    insights_max_retries: int = 3
    """The max number of retries allowed for a single message."""


from abc import ABC
from dataclasses import dataclass, field
import time
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.message import Message
from loguru import logger as log

@dataclass
class BlackListEntry(ABC):
    reason: str = None
    date: int = field(default_factory=lambda: int(time.time()))
    ttl: int = None

class InMemBlackListMiddleware:
    """Middleware to block users based on a blacklist. The blacklist is stored in memory."""
    
    def __init__(self, 
                 black_list: dict[str, BlackListEntry] = None,
                 reject_message: str = "You are banned from using this service"):
        # session_id -> BlackListEntry
        self.black_list = black_list or {}
        self.reject_message = reject_message
        
    async def __call__(self, message: Message, connector: BaseConnector, assistant: BaseAssistant):
        assert isinstance(message, Message), "Message must be a Message object"

        id =  message.lead.get_session_id()
        source = message.lead.connector_name
        entry = self.get_entry(id)
        if entry:
            log.critical(f"User {id} from {source} is blacklisted. Reason: {entry.reason}")
            if connector:
                await connector.send_text_message(message.lead, self.reject_message)
            return False
        else:
            return True
    
    def add_to_black_list(self, id: str, reason: str = None, ttl: int = None):
        self.black_list[id] = BlackListEntry(reason=reason, date=int(time.time()), ttl=ttl)

    def remove_from_black_list(self, id: str):
        self.black_list.pop(id, None)
        
    def get_entry(self, id: str):
        entry = self.black_list.get(id)
        # check that the entry is not expired
        if entry and entry.ttl:
            if entry.date + entry.ttl < time.time():
                self.remove_from_black_list(id)
                return None

        return entry
    


import json
from loguru import logger as log
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.message import Message
from cel.gateway.model.attachment import ContactAttachment



class ContactDecodingMiddleware:
    
    
    def __init__(self, 
                 msg_prefix: str = "User shared a contact: ",
                 on_fail_message: str = 'User shared a contact but it could not be decoded.'):
        log.debug(f"ContactDecodingMiddleware initialized")
        self.msg_prefix = msg_prefix
        self.on_fail_message = on_fail_message

        
    async def __call__(self, message: Message, connector: BaseConnector, assistant: BaseAssistant):
        assert isinstance(message, Message), "Message must be a Message object"
        
        try:
            
            
            if not message.attachments or len(message.attachments) < 1:
                return True
            
            # Get first attachment
            attach = message.attachments[0]
            
            if attach.type != "contact":
                log.error(f"Attachment is not a voice message")
                return True
            
            log.debug(f"Message {message.lead.get_session_id()} -> Decoding contact attachment")
            
            assert isinstance(attach, ContactAttachment), "Attachment must be a ContactAttachment"
            
            # Decode contact attachment
            cto_metadata = attach.metadata
            cto_name = attach.name
            
            # Build contact text
            cto_data = json.dumps(cto_metadata)
            text = f"{self.msg_prefix} Name: {cto_name}. Data: {cto_data}"
            
            # if text is empty or None, set the on_fail_message
            message.text = text if text else self.on_fail_message
            
            return True
        except Exception as e:
            log.error(f"Failed to decode contact attachment: {e}")
            # return True, so fails on STT won't stop the message processing
            if self.on_fail_message:
                message.text = self.on_fail_message
            return True
        
    




from .component import Component



#  {
#     "type": "text",
#     "text": "replacement_text"
#   }
class Text(Component):
    def __init__(self, text: str) -> None:
        self.text = text

    def __str__(self):
        return f'{{"type": "text", "text": "{self.text}"}}'

    def __repr__(self):
        return f"TextComponent(text={self.text})"

    def __eq__(self, other):
        return self.text == other.text
    

if __name__ == "__main__":
    text = Text("replacement_text")
    print(text)
    # Expected output: '{"type": "text", "text": "replacement_text"}'




import json
from typing import Any
from loguru import logger as log
from cel.gateway.model.conversation_lead import ConversationLead
from cel.stores.state.base_state_provider import BaseChatStateProvider
      

class AsyncStateManager:

    def __init__(self, 
                 lead: ConversationLead,
                 store: BaseChatStateProvider,
                 commit_on_error: bool = False
                ):
        """ StateManager is a context manager to manage the state of a conversation.
        
        :param lead: ConversationLead
        :param store: BaseChatStateProvider
        :param commit_on_error: if True, the state will be saved on error, default is False
        """
        
        if not isinstance(store, BaseChatStateProvider):
            raise ValueError("StateManager: store must be an instance of BaseChatStateProvider")
        if not isinstance(lead, ConversationLead):
            raise ValueError("StateManager: lead must be an instance of ConversationLead")
        
        self.store = store
        self.lead = lead
        self.state = {}
        self.commit_on_error = commit_on_error
        

    async def __aenter__(self):
        await self.load_state()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            log.error(f"StateManager: error on context execution {exc_type} {exc_value}")
            if self.commit_on_error:
                await self.save_state()
            # Return False to raise the exception
            return False
        else:
            await self.save_state()
            return True


    async def load_state(self):
        self.state = await self.store.get_store(self.lead.get_session_id()) or {}
        return self.state

    async def save_state(self):
        await self.store.set_store(self.lead.get_session_id(), self.state)
        
    def get(self, key, default=None):
        return self.state.get(key, default)
    
    def set(self, key, value):
        self.state[key] = value

    def __getitem__(self, key):
        return self.state.get(key)

    def __setitem__(self, key, value):
        self.state[key] = value







from abc import abstractmethod
from loguru import logger as log
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.message import Message
from cel.gateway.model.message_gateway_context import MessageGatewayContext
from cel.gateway.model.outgoing.outgoing_message import OutgoingMessage


class BaseMiddleware:

    @abstractmethod
    async def incoming_message(self,  
                               message: Message, 
                               connector: BaseConnector, 
                               assistant: BaseAssistant):
        
        raise NotImplementedError
    
    
    
    @abstractmethod
    async def outgoing_message(self, 
                               message: OutgoingMessage, 
                               connector: BaseConnector, 
                               assistant: BaseAssistant, 
                               is_partial: bool = False, 
                               is_summary: bool = False,
                               mode = None):
        
        raise NotImplementedError
    
    
    
    @abstractmethod
    async def startup(self, ctx: MessageGatewayContext):
        pass


import asyncio
from typing import Callable
import pysbd

from cel.assistants.stream_content_chunk import StreamContentChunk


def streaming_sentence_detector(stream, language="en"):
    """pySBD based streaming sentence detector"""
    
    buffer = ''
    seg = pysbd.Segmenter(language=language, clean=False)
    for char in stream:
        buffer += char
        sentences = seg.segment(buffer)
        if len(sentences) > 1:
            yield from sentences[:-1]
            buffer = sentences[-1]
    if buffer:
        yield buffer
        
async def streaming_word_detector_async(stream):
    def pop_word(s):
        if s:
            word = s.split(' ')[0]
            s = s[len(word):]
            return word, s
        return None, s

    buffer = ""
    async for token in stream:
        buffer += token
        word, buffer = pop_word(buffer)
        if word:
            yield word
            
    if buffer:
        yield buffer


async def streaming_sentence_detector_async(stream, language="en", on_chunk = None):
    """pySBD based streaming sentence detector"""
    
    buffer = ''
    seg = pysbd.Segmenter(language=language, clean=False)
    async for chunk in stream:
        # cast char to class StreamChunk
        assert isinstance(chunk, StreamContentChunk), "stream must be a StreamChunk"
      
        if not chunk.is_partial:
            yield StreamContentChunk(content=chunk.content, is_partial=False)     
            break
      
        buffer += chunk.content   
        
        if on_chunk:
            await on_chunk(chunk, buffer)
            
        sentences = seg.segment(buffer)
        if len(sentences) > 1:
            for sentence in sentences[:-1]:
                yield StreamContentChunk(content=sentence, is_partial=chunk.is_partial)
            buffer = sentences[-1]
    if buffer:
        yield StreamContentChunk(content=buffer, is_partial=chunk.is_partial)





if __name__ == "__main__":

    text = """Holi is a popular ancient Hindu festival https://github.com/nipunsadvilkar/pySBD, also known as the "Festival of Colors" or the "Festival of Love." The festival celebrates the arrival of spring, the end of winter, the blossoming of love, and for many it's a festive day to meet others, play and laugh, forget and forgive, and repair broken relationships. It is also celebrated as a thanksgiving for a good harvest. Holi is typically celebrated in India, Nepal, and other parts of the world with significant populations of Hindus or people of Indian origin. The festival has spread to parts of Europe and North America as a spring celebration of love, frolic, and colors. Holi celebrations start on the night before Holi with a Holika Dahan where people gather, perform religious rituals in front of the bonfire, and pray that their internal evil be destroyed the way Holika, the sister of the demon king Hiranyakashipu, was killed in the fire. The next morning is celebrated as Rangwali Holi – a free-for-all festival of colors, where participants play, chase and color each other with dry powder and colored water, with some carrying water guns and colored water-filled balloons for their water fight. The festival also provides an opportunity to enjoy seasonal foods and delicacies, socialize, and enjoy festive activities. While Holi is an ancient festival in India and Nepal, it is now celebrated in many places across the world."""    

    # #  SYNC TEST
    # ---------------------------------------------------------------------
    # text_stream = iter(text)
    
    
    
    # for sentence in streaming_sentence_detector(text_stream):
    #     print(sentence)
    #     print ("----------------------")


    # print ("-----------DONE-----------")
    # print ("-----------DONE-----------")


    # ASYNC TEST
    # ---------------------------------------------------------------------
    class AsyncIterable:
        def __init__(self, iterable):
            self.iterable = iterable

        def __aiter__(self):
            return self

        async def __anext__(self):
            try:
                value = StreamContentChunk(content=next(self.iterable), is_partial=True)
            except StopIteration:
                raise StopAsyncIteration
            await asyncio.sleep(0)  # This makes this method asynchronous
            return value


    async def async_test():
        # ASYNC
        import asyncio
        text_stream = AsyncIterable(iter(text))
        
        count = 0
        async for sentence in streaming_sentence_detector_async(text_stream):
            print(f"Sentence {count}::: {sentence.content}")
            count +=1
            print ("----------------------")        
        
        

    asyncio.run(async_test())
    


from aiogram import Bot
from cel.gateway.model.attachment import FileAttachment,\
                                                LocationAttachment,\
                                                MessageAttachmentType

class TelegramLocationAttachment(LocationAttachment):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)



#  add this fields to the class:
# fileSize: int;
# width: int;
# height: int;   
class TelegramAttachment(FileAttachment):
    
    def __init__(self, 
                 type: MessageAttachmentType = None,
                 title: str = None, 
                 description: str = None, 
                 mimeType: str = None, 
                 metadata: any = None, 
                 fileSize: int = None, 
                 width: int = None, 
                 height: int = None,
                 file_url: str = None
                ):
        super().__init__(title, description, mimeType, metadata, type=type, file_url=file_url)
        self.fileSize: int = fileSize
        self.width: int = width
        self.height: int = height
    
    def to_dict(self):
        data = super().to_dict()
        data['fileSize'] = self.fileSize
        data['width'] = self.width
        data['height'] = self.height
        return data
    
    @classmethod
    def from_dict(cls, attachment_dict):
        return TelegramAttachment(
            title=attachment_dict.get("title"),
            description=attachment_dict.get("description"),
            content=attachment_dict.get("content"),
            mimeType=attachment_dict.get("mimeType"),
            metadata=attachment_dict.get("metadata"),
            fileSize=attachment_dict.get("fileSize"),
            width=attachment_dict.get("width"),
            height=attachment_dict.get("height")
        )
    
    def __str__(self):
        return f"TelegramAttachment: {self.title}"
    
    def __repr__(self):
        return f"TelegramAttachment: {self.title}"
    
    
    @classmethod
    async def load_from_message(cls, message: dict, token: str):
        msg = message.get("message")
        # check if the message has a photo
        if 'photo' in msg:
            return await cls.load_image_from_message(message, token)
        if 'voice' in msg:
            return await cls.load_audio_from_message(message, token)
        if 'location' in msg:
            return await cls.load_location_from_message(message)
        
        
    @classmethod
    async def load_location_from_message(cls, message: dict):
        msg = message.get("message")
        location = msg.get('location')
        latitude = location['latitude']
        longitude = location['longitude']
        metadata = {}
        description = ""
        if 'venue' in msg:
            venue = msg['venue']
            metadata['title'] = venue.get('title')
            metadata['address'] = venue.get('address')
            metadata['foursquare_id'] = venue.get('foursquare_id')
            metadata['foursquare_type'] = venue.get('foursquare_type')  
            description = venue['title'] + " at " + venue['address'] + "(" + venue.get('foursquare_type') + ")"
        
        return TelegramLocationAttachment(latitude=latitude, 
                                          longitude=longitude,
                                          description=description,
                                          metadata=metadata)
        
    @classmethod
    async def load_audio_from_message(cls, message: dict, token: str):
        
        msg = message.get("message")
        audio = msg['voice']
        file_id = audio['file_id']
        file_unique_id = audio['file_unique_id']
        file_size = audio['file_size']
        duration = audio['duration']
        mime_type = audio['mime_type']
        metadata = {
            'file_id': file_id,
            'file_unique_id': file_unique_id,
            'file_size': file_size,
            'duration': duration,
            'mime_type': mime_type
        }
        
        # get file url
        file_url = "https://es.wikipedia.org/static/images/icons/wikipedia.png"
        if file_id != '123':
            bot = Bot(token=token)
            file = await bot.get_file(file_id)
            file_url = f"https://api.telegram.org/file/bot{token}/{file.file_path}" 
        
        return TelegramAttachment(
            type=MessageAttachmentType.VOICE,
            title="audio",
            description="audio",
            mimeType=mime_type,
            metadata=metadata,
            fileSize=file_size,
            width=None,
            height=None,
            file_url=file_url
        )


    @classmethod
    async def load_image_from_message(cls, message: dict, token: str):
        
        msg = message.get("message")
        
        photo = msg['photo'][-1]
        file_id = photo['file_id']
        file_unique_id = photo['file_unique_id']
        file_size = photo['file_size']
        width = photo['width']
        height = photo['height']
        metadata = {
            'file_id': file_id,
            'file_unique_id': file_unique_id,
            'file_size': file_size,
            'width': width,
            'height': height
        }
        
        # get file url
        file_url = f"https://es.wikipedia.org/static/images/icons/wikipedia.png"
        if file_id != '123':
            bot = Bot(token=token)
            file = await bot.get_file(file_id)
            file_url = file.file_path
        
        return TelegramAttachment(
            type=MessageAttachmentType.IMAGE,
            title="photo",
            description="photo",
            mimeType="image/jpeg",
            metadata=metadata,
            fileSize=file_size,
            width=width,
            height=height,
            file_url=file_url
        )        
    
    


from abc import ABC
import asyncio
import threading
import time
from typing import Callable
from openai import OpenAI
from openai.types.moderation import Moderation
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.message import Message
from loguru import logger as log
from dataclasses import dataclass

from cel.middlewares.moderation.moderation_events import ModMiddlewareEvents

@dataclass
class RedFlagEntry(ABC):
    count: int = 0
    updated_at: int = 0
    

class OpenAIEndpointModerationMiddleware():
    """ OpenAIEndpointModerationMiddleware is a middleware that uses OpenAI API to moderate messages.
    It uses the OpenAI API to moderate messages and flags them if they are inappropriate.
    The middleware bubbles up the flagged messages to further processing using the `on_message_flagged` event.
    
    Events:
     - on_message_flagged: Event that is triggered when a message is flagged by the OpenAI API.
     
    Args:
        - custom_evaluation_function: A custom function that accepts a message and returns a Moderation object.
        - on_mod_fail_continue: A boolean that determines if the middleware should continue processing if the moderation fails.
        - expire_after: An integer that determines the time in seconds after which the user flags should be reset.
        
        For more info follow https://platform.openai.com/docs/guides/moderation
    """
    
    def __init__(self,
                 custom_evaluation_function: Callable[[str], Moderation] = None,
                 enable_expiration: bool = False,
                 expire_after: int = 86400,
                 prunning_interval: int = 60,
                 on_mod_fail_continue: bool = True):
        
        self.custom_evaluation_function = custom_evaluation_function
        self.on_mod_fail_continue = on_mod_fail_continue
        self.client = OpenAI()
        self.user_flags = {}
        self.expire_after = expire_after
        self.prunning_interval = prunning_interval
        
        
        if enable_expiration:
            # Asegúrate de que el bucle de eventos esté corriendo
            self.loop = asyncio.get_event_loop()
            if not self.loop.is_running():
                threading.Thread(target=self._start_event_loop, daemon=True).start()
            
            # Inicia la coroutine en segundo plano
            self.loop.call_soon_threadsafe(asyncio.create_task, self.reset_user_flags_loop()) 
                           
        
    def _start_event_loop(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()
        
        
    async def reset_user_flags_loop(self):
        while True:            
            await asyncio.sleep(self.prunning_interval)
            log.debug(f"OpenAIEndpointModerationMiddleware: Prunning user flags each {self.prunning_interval} seconds")
            current_time = time.time()
            expired_sessions = [
                session_id for session_id, flags in self.user_flags.items()
                if current_time - flags.updated_at > self.expire_after
            ]

            for session_id in expired_sessions:
                self.reset_user_flags(session_id)
        

    async def __call__(self, message: Message, connector: BaseConnector, assistant: BaseAssistant):
        
        try:
            text = message.text
            log.debug(f"OpenAIEndpointModerationMiddleware: {text}")
            
            moderation = self.client.moderations.create(input=text)
            result = moderation.results[0]
            assert isinstance(result, Moderation)
            
            log.debug(f"OpenAIEndpointModerationMiddleware: {text} -> {result.flagged}")
            
            custom_flagged = False
            if self.custom_evaluation_function:
                custom_flagged = self.custom_evaluation_function(result)
            
            if result.flagged or custom_flagged:
                session_id = message.lead.get_session_id()
                count = await self.__count_flagged(session_id)
                # append to message metadata moderation result, init metadata if not exists
                report = {
                    'flagged': result.flagged,
                    'count': count,
                    custom_flagged: custom_flagged,
                    'results': result
                }
                message.metadata = message.metadata or {}
                message.metadata['moderation'] = report
                
                # (self, event_name: str, lead: ConversationLead, message: Message = None, connector: BaseConnector = None, data: dict= None) -> EventResponse:
                await assistant.call_event(
                    ModMiddlewareEvents.on_message_flagged, 
                    lead=message.lead, 
                    connector=connector,
                    data=report)
                
            return True
            
        except Exception as e:
            log.error(f"OpenAIEndpointModerationMiddleware: {e}")
            if self.on_mod_fail_continue:
                return True
            else:
                return False
            
            
    async def __count_flagged(self, session_id: str):
        if session_id in self.user_flags:
            self.user_flags[session_id].count += 1
        else:
            self.user_flags[session_id] = RedFlagEntry(count=1, updated_at=time.time())
            # self.user_flags[session_id].count = 1
            # self.user_flags[session_id].updated_at = time.time()
            
        return self.user_flags[session_id].count
    
    
    def reset_user_flags(self, session_id: str):
        if session_id in self.user_flags:
            del self.user_flags[session_id]
        return True
    
    
    def get_user_flags(self, session_id: str):
        if session_id in self.user_flags:
            return self.user_flags[session_id]
        return None


"""
# Llama3Guard Moderation Documentation

## Overview

Llama3Guard Moderation is a component designed to detect and classify toxic messages within the Cel.ai platform. It leverages the Llama Guard 3 model, which is fine-tuned for content safety classification. This document provides an overview of how Llama3Guard Moderation works, including its hazard taxonomy, supported languages, and example usage.

## Model Details

Llama Guard 3 is a Llama-3.1-8B pretrained model, fine-tuned for content safety classification. It can classify content in both LLM inputs (prompt classification) and LLM responses (response classification). The model generates text indicating whether a given prompt or response is safe or unsafe. If unsafe, it also lists the content categories violated.

### Hazard Taxonomy and Policy

The model is trained to predict safety labels based on the MLCommons taxonomy of 13 hazards, plus an additional category for Code Interpreter Abuse. The categories are as follows:

1. **Violent Crimes**: Responses that enable, encourage, or endorse violent crimes, including unlawful violence toward people and animals.
2. **Non-Violent Crimes**: Responses that enable, encourage, or endorse non-violent crimes, including personal crimes, financial crimes, property crimes, drug crimes, weapons crimes, and cyber crimes.
3. **Sex-Related Crimes**: Responses that enable, encourage, or endorse sex-related crimes, including sex trafficking, sexual assault, sexual harassment, and prostitution.
4. **Child Sexual Exploitation**: Responses that contain, describe, enable, encourage, or endorse the sexual abuse of children.
5. **Defamation**: Responses that are both verifiably false and likely to injure a living person’s reputation.
6. **Specialized Advice**: Responses that contain specialized financial, medical, or legal advice, or that indicate dangerous activities or objects are safe.
7. **Privacy**: Responses that contain sensitive, nonpublic personal information that could undermine someone’s physical, digital, or financial security.
8. **Intellectual Property**: Responses that may violate the intellectual property rights of any third party.
9. **Indiscriminate Weapons**: Responses that enable, encourage, or endorse the creation of indiscriminate weapons, including chemical, biological, radiological, nuclear, and high-yield explosive weapons.
10. **Hate**: Responses that demean or dehumanize people based on their sensitive, personal characteristics.
11. **Suicide & Self-Harm**: Responses that enable, encourage, or endorse acts of intentional self-harm, including suicide, self-injury, and disordered eating.
12. **Sexual Content**: Responses that contain erotica.
13. **Elections**: Responses that contain factually incorrect information about electoral systems and processes.
14. **Code Interpreter Abuse**: Responses that seek to abuse code interpreters, including those that enable denial of service attacks, container escapes, or privilege escalation exploits.

### Supported Languages

Llama Guard 3 supports content safety for the following languages:
- English
- French
- German
- Hindi
- Italian
- Portuguese
- Spanish
- Thai

In this example, the `classify` function takes a response as input and returns a classification indicating whether the response is safe or unsafe. If unsafe, it also lists the violated categories.

## Conclusion

Llama3Guard Moderation is a powerful tool for detecting and classifying toxic messages within the Cel.ai platform. By leveraging the Llama Guard 3 model, it ensures content safety across multiple languages and a comprehensive hazard taxonomy.
"""

from abc import ABC
import asyncio
from dataclasses import dataclass
import os
import threading
import time
from typing import Callable
from together import Together
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.message import Message
from loguru import logger as log

from cel.middlewares.moderation.moderation_events import ModMiddlewareEvents

hazard_categories = {
    "S1": "Violent Crimes",
    "S2": "Non-Violent Crimes",
    "S3": "Sex-Related Crimes",
    "S4": "Child Sexual Exploitation",
    "S5": "Defamation",
    "S6": "Specialized Advice",
    "S7": "Privacy",
    "S8": "Intellectual Property",
    "S9": "Indiscriminate Weapons",
    "S10": "Hate",
    "S11": "Suicide & Self-Harm",
    "S12": "Sexual Content",
    "S13": "Elections",
    "S14": "Code Interpreter Abuse"
}

class RedFlagEntry(ABC):
    count: int = 0
    updated_at: int = 0

@dataclass
class ModerationResult(ABC):
    flagged: bool = False
    category: str = None
    description: str = None


class Llama3GuardModerationMiddleware():
    """ OpenAIEndpointModerationMiddleware is a middleware that uses OpenAI API to moderate messages.
    It uses Llama3.1 Guard 8B model to moderate messages and flags them if they are inappropriate.
    The middleware bubbles up the flagged messages to further processing using the `on_message_flagged` event.
    
    You must provide the Together API key in the environment variable:
        - TOGETHER_API_KEY: The Together API key for the assistant.
    
    Events:
        - on_message_flagged: Event that is triggered when a message is flagged by the OpenAI API.
     
    Args:
        - custom_evaluation_function: A custom function that accepts a message and returns a Moderation object.
        - on_mod_fail_continue: A boolean that determines if the middleware should continue processing if the moderation fails.
        - expire_after: An integer that determines the time in seconds after which the user flags should be reset.
        
        For more info follow https://huggingface.co/meta-llama/Llama-Guard-3-8B
    """
    
    def __init__(self,
                 custom_evaluation_function: Callable[[str], ModerationResult] = None,
                 enable_expiration: bool = False,
                 expire_after: int = 86400,
                 prunning_interval: int = 60,
                 on_mod_fail_continue: bool = True):
        
        self.custom_evaluation_function = custom_evaluation_function
        self.on_mod_fail_continue = on_mod_fail_continue
        
        assert os.environ.get('TOGETHER_API_KEY'), "TOGETHER_API_KEY is not set in the environment variables"
        self.client = Together(api_key=os.environ.get('TOGETHER_API_KEY'))
        
        self.user_flags = {}
        self.expire_after = expire_after
        self.prunning_interval = prunning_interval
        
        if enable_expiration:
            # Asegúrate de que el bucle de eventos esté corriendo
            self.loop = asyncio.get_event_loop()
            if not self.loop.is_running():
                threading.Thread(target=self._start_event_loop, daemon=True).start()
            
            # Inicia la coroutine en segundo plano
            self.loop.call_soon_threadsafe(asyncio.create_task, self.reset_user_flags_loop()) 
                           
        
    def _start_event_loop(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()
        
        
        
    async def reset_user_flags_loop(self):
        while True:
            await asyncio.sleep(self.prunning_interval)
            log.debug(f"Llama3GuardModerationMiddleware: Prunning user flags each {self.prunning_interval} seconds")
            current_time = time.time()
            expired_sessions = [
                session_id for session_id, flags in self.user_flags.items()
                if current_time - flags.updated_at > self.expire_after
            ]

            for session_id in expired_sessions:
                self.reset_user_flags(session_id)
        
        
        
        
    async def evaluate(self, text: str):
        response = self.client.chat.completions.create(
            model="meta-llama/Meta-Llama-Guard-3-8B",
            messages=[{"role": "user", "content": text}],
        )        
        res = response.choices[0].message.content
        # res = "safe" then flagged = False
        # if res = "unsafe" then flagged = True
        # when res = "unsafe" the next line will be the category of the hazard
        # for example:
        #  "unsafe"
        #  "S12"
        if res.startswith("unsafe"):
            category = res.split("\n")[1]
            desc=hazard_categories[category]
            return ModerationResult(flagged=True, category=category, description=desc)
        else:
            return ModerationResult(flagged=False)



    async def __call__(self, message: Message, connector: BaseConnector, assistant: BaseAssistant):
        
        try:
            text = message.text
            log.debug(f"Llama3GuardModerationMiddleware: {text}")
            
            result = await self.evaluate(text)
            assert isinstance(result, ModerationResult)
            
            log.debug(f"Llama3GuardModerationMiddleware: {text} -> {result.flagged}")
            
            custom_flagged = False
            if self.custom_evaluation_function:
                custom_flagged = self.custom_evaluation_function(result)
            
            if result.flagged or custom_flagged:
                session_id = message.lead.get_session_id()
                count = await self.__count_flagged(session_id)
                # append to message metadata moderation result, init metadata if not exists
                report = {
                    'flagged': result.flagged,
                    'count': count,
                    'custom_flagged': custom_flagged,
                    'results': result
                }
                message.metadata = message.metadata or {}
                message.metadata['moderation'] = report
                
                # (self, event_name: str, lead: ConversationLead, message: Message = None, connector: BaseConnector = None, data: dict= None) -> EventResponse:
                await assistant.call_event(
                    ModMiddlewareEvents.on_message_flagged, 
                    lead=message.lead, 
                    connector=connector,
                    data=report)
                
            return True
            
        except Exception as e:
            log.error(f"Llama3GuardModerationMiddleware: {e}")
            if self.on_mod_fail_continue:
                return True
            else:
                return False
            
            
    async def __count_flagged(self, session_id: str):
        if session_id in self.user_flags:
            self.user_flags[session_id].count += 1
        else:
            self.user_flags[session_id] = RedFlagEntry()
            self.user_flags[session_id].count = 1
            self.user_flags[session_id].updated_at = time.time()
            
        return self.user_flags[session_id].count
    
    
    def reset_user_flags(self, session_id: str):
        if session_id in self.user_flags:
            del self.user_flags[session_id]
        return True
    
    
    def get_user_flags(self, session_id: str):
        if session_id in self.user_flags:
            return self.user_flags[session_id]
        return None
    



from aiogram import Bot
from loguru import logger as log
from cel.connectors.whatsapp.model.media_utils import query_media_url
from cel.gateway.model.attachment import FileAttachment, \
                                        LocationAttachment, \
                                        MessageAttachmentType, \
                                        ContactAttachment


sample_data = {
  "object": "whatsapp_business_account",
  "entry": [
    {
      "id": "11232312744885411",
      "changes": [
        {
          "value": {
            "messaging_product": "whatsapp",
            "metadata": {
              "display_phone_number": "15553453411",
              "phone_number_id": "16563546234234211"
            },
            "contacts": [
              {
                "profile": { "name": "Ruben Sarasa" },
                "wa_id": "52161231257911"
              }
            ],
            "messages": [
              {
                "from": "52161231257911",
                "id": "wamid.SDFAASDASDSADS==",
                "timestamp": "1746830741",
                "type": "contacts",
                "contacts": [
                  {
                    "name": {
                      "first_name": "Jonathan",
                      "last_name": "Doe",
                      "formatted_name": "Jonathan Doe"
                    },
                    "phones": [
                      {
                        "phone": "+1 123 205 1111",
                        "wa_id": "5216621111",
                        "type": "CELL"
                      }
                    ]
                  }
                ]
              }
            ]
          },
          "field": "messages"
        }
      ]
    }
  ]
}


class WhatsappAttachment(FileAttachment):

    def __init__(self,  
                 type: MessageAttachmentType = None,
                 title: str = None,
                 description: str = None,
                 mimeType: str = None,
                 metadata: any = None,
                 fileSize: int = None,
                 width: int = None,
                 height: int = None,
                 file_url: str = None,
                 caption: str = None):
        
        super().__init__(title, description, mimeType=mimeType, metadata=metadata, type=type)
        self.fileSize: int = fileSize
        self.width: int = width
        self.height: int = height
        assert isinstance(file_url, str), "file_url must be a string"
        self.file_url = file_url
        # self.file_url: str = file_url,
        self.caption: str = caption
    
    def to_dict(self):
        data = super().to_dict()
        data['fileSize'] = self.fileSize
        data['width'] = self.width
        data['height'] = self.height
        data['file_url'] = self.file_url
        return data
    
    @classmethod
    def from_dict(cls, attachment_dict):
        return WhatsappAttachment(
            title=attachment_dict.get("title"),
            description=attachment_dict.get("description"),
            content=attachment_dict.get("content"),
            mimeType=attachment_dict.get("mimeType"),
            metadata=attachment_dict.get("metadata"),
            fileSize=attachment_dict.get("fileSize"),
            width=attachment_dict.get("width"),
            height=attachment_dict.get("height"),
            file_url=attachment_dict.get("file_url")
        )
    
    def __str__(self):
        return f"WhatsappAttachment: {self.title}"
    
    def __repr__(self):
        return f"WhatsappAttachment: {self.title}"
    
    
    @classmethod
    async def load_from_message(cls, data: dict, token: str, phone_number_id: str):
        assert isinstance(data, dict), "data must be a dictionary"
        msg = data.get("entry")[0].get("changes")[0].get("value").get("messages")[0]

        # check if the message has a photo
        if 'image' in msg:
            log.info("Loading image from message")
            return await cls.load_image_from_message(data, token, phone_number_id)
        if 'contacts' in msg:
            log.info("Loading contact from message")
            return await cls.load_contact_from_message(data)
            
        # if 'voice' in msg:
        #     return await cls.load_audio_from_message(message, token)
        # if 'location' in msg:
        #     return await cls.load_location_from_message(message)
        
    @classmethod
    async def load_image_from_message(cls, data: dict, token: str, phone_number_id: str):
        
        msg = data.get("entry")[0].get("changes")[0].get("value").get("messages")[0]
        image = msg.get('image')
        
        caption = image.get('caption', '')
        mime_type = image.get('mime_type', 'image/jpeg')
        file_id = image.get('id')
        sha256 = image.get('sha256')
        file_url = query_media_url(file_id, token)
        metadata = {
            'sha256': sha256,
            'file_id': file_id,
            'caption': caption,
            'phone_number_id': phone_number_id,
            'token': token
        }
        
        return WhatsappAttachment(
            type=MessageAttachmentType.IMAGE,
            title=caption,
            description="photo",
            file_url=file_url,
            mimeType=mime_type,
            metadata=metadata,
            fileSize=None,
            width=None,
            height=None
        )        
    

    @classmethod
    async def load_contact_from_message(cls, data: dict):
        msg = data.get("entry")[0].get("changes")[0].get("value").get("messages")[0]
        contact = msg.get('contacts')[0]
        name = contact.get('name', {}).get('formatted_name', '')
        metadata = contact        

        return ContactAttachment(
            name=name,
            metadata=metadata
        )


import aiohttp
from loguru import logger as log
from typing import Optional, Dict
from loguru import logger as log    
from typing import Any, Dict, Optional


class ChatwootClient:

    def __init__(self,
                 base_url: str,
                 account_id: str,
                 access_key: str,
                 headers: Optional[Dict[str, str]] = None,
                 ssl: bool = False):
        self.base_url = base_url
        self.account_id = account_id
        self.access_key = access_key
        self.headers = headers or {}
        self.headers.update({
            'api_access_token': access_key
        })
        self.ssl = ssl

    async def list_agent_bots(self) -> Dict[str, Any]:
        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/agent_bots"
        log.debug(f"Listing agent bots from Chatwoot url: {url}")

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.get(url, headers=self.headers) as response:
                response_data = await response.json()
                return response_data


    async def create_contact(self,
                                inbox_id: int,
                                name: Optional[str] = None,
                                email: Optional[str] = None,
                                phone_number: Optional[str] = None,
                                avatar: Optional[str] = None,
                                avatar_url: Optional[str] = None,
                                identifier: Optional[str] = None,
                                custom_attributes: Optional[Dict[str, str]] = None
                            ) -> Dict[str, Any]:

        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/contacts"
        log.debug(f"Creating contact at Chatwoot url: {url}")

        payload = {
            'inbox_id': inbox_id,
            'name': name,
            'email': email,
            'phone_number': phone_number,
            'avatar': avatar,
            'avatar_url': avatar_url,
            'identifier': identifier,
            'custom_attributes': custom_attributes
        }

        payload = {k: v for k, v in payload.items() if v is not None}

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.post(url, json=payload, headers=self.headers) as response:
                response_data = await response.json()
                return response_data
            
    async def update_contact(self,
                                contact_id: int,
                                inbox_id: Optional[int] = None,
                                name: Optional[str] = None,
                                email: Optional[str] = None,
                                phone_number: Optional[str] = None,
                                avatar: Optional[str] = None,
                                avatar_url: Optional[str] = None,
                                identifier: Optional[str] = None,
                                custom_attributes: Optional[Dict[str, str]] = None
                            ) -> Dict[str, Any]:
        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/contacts/{contact_id}"
        log.debug(f"Updating contact at Chatwoot url: {url}")

        payload = {
            'inbox_id': inbox_id,
            'name': name,
            'email': email,
            'phone_number': phone_number,
            'avatar': avatar,
            'avatar_url': avatar_url,
            'identifier': identifier,
            'custom_attributes': custom_attributes
        }

        payload = {k: v for k, v in payload.items() if v is not None}

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.put(url, json=payload, headers=self.headers) as response:
                response_data = await response.json()
                return response_data
            
    async def get_contact(self, contact_id: int) -> Dict[str, Any]:
        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/contacts/{contact_id}"
        log.debug(f"Getting contact from Chatwoot url: {url}")
        
        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.get(url, headers=self.headers) as response:
                response_data = await response.json()
                return response_data
    
            


    # GET https://app.chatwoot.com/api/v1/accounts/{account_id}/contacts/search?q=identifier:1234567890            
    async def search_contact(self, query: str) -> Dict[str, Any]:
        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/contacts/search"
        log.debug(f"Searching contact at Chatwoot url: {url}")
        url = f"{url}?q={query}"

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.get(url, headers=self.headers) as response:
                response_data = await response.json()
                return response_data



    async def create_new_conversation(self,
                                        source_id: str,
                                        inbox_id: int,
                                        contact_id: int,
                                        additional_attributes: Optional[Dict[str, str]] = None,
                                        custom_attributes: Optional[Dict[str, str]] = None,
                                        status: Optional[str] = "open",
                                        assignee_id: Optional[str] = None,
                                        team_id: Optional[str] = None,
                                        message: Optional[Dict[str, str]] = None
                                    ) -> Dict[str, Any]:

        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/conversations"
        log.debug(f"Creating conversation at Chatwoot url: {url}")

        payload = {
            'source_id': source_id,
            'inbox_id': inbox_id,
            'contact_id': contact_id,
            'additional_attributes': additional_attributes,
            'custom_attributes': custom_attributes,
            'status': status,
            'assignee_id': assignee_id,
            'team_id': team_id,
            'message': message
        }

        payload = {k: v for k, v in payload.items() if v is not None}

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.post(url, json=payload, headers=self.headers) as response:
                response_data = await response.json()
                return response_data

    async def create_message(self,
                                account_id: int,
                                conversation_id: int,
                                content: str,
                                message_type: Optional[str] = "outgoing",
                                private: Optional[bool] = False,
                                content_type: Optional[str] = "text",
                                content_attributes: Optional[Dict[str, str]] = None,
                                template_params: Optional[Dict[str, str]] = None
                            ) -> Dict[str, Any]:
        """ Create a new message in Chatwoot 

            content_type: "text", "input_email", "cards", "input_select", "form", "article"
        """
        # https://app.chatwoot.com/api/v1/accounts/{account_id}/conversations/{conversation_id}/messages
        url = f"{self.base_url}/api/v1/accounts/{account_id}/conversations/{conversation_id}/messages"
        log.debug(f"Creating message at Chatwoot url: {url}")

        payload = {
            'content': content,
            'message_type': message_type,
            'private': private,
            'content_type': content_type,
            'content_attributes': content_attributes,
            'template_params': template_params
        }

        payload = {k: v for k, v in payload.items() if v is not None}

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.post(url, json=payload, headers=self.headers) as response:
                response_data = await response.json()
                return response_data

    async def get_inbox(self, account_id: int, inbox_id: int) -> Dict[str, Any]:
        url = f"{self.base_url}/api/v1/accounts/{account_id}/inboxes/{inbox_id}"
        log.debug(f"Getting inbox from Chatwoot url: {url}")

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.get(url, headers=self.headers) as response:
                response_data = await response.json()
                return response_data

    async def get_inboxes(self, account_id: int) -> Dict[str, Any]:
        url = f"{self.base_url}/api/v1/accounts/{account_id}/inboxes"
        log.debug(f"Getting inboxes from Chatwoot url: {url}")

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.get(url, headers=self.headers) as response:
                response_data = await response.json()
                return response_data

    async def get_inbox_by_name(self, account_id: int, name: str) -> Dict[str, Any]:
        payload = await self.get_inboxes(account_id)
        inboxes = payload.get("payload", {})
        for inbox in inboxes:
            if inbox.get("name") == name:
                return inbox

        return None

    async def create_api_inbox(self, account_id: int, name: str, webhook_url: str) -> Dict[str, Any]:
        url = f"{self.base_url}/api/v1/accounts/{account_id}/inboxes"
        log.debug(f"Creating inbox at Chatwoot url: {url}")

        payload = {
            'name': name,
            'channel': {
                'type': 'api',
                'webhook_url': webhook_url
            }
        }

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.post(url, json=payload, headers=self.headers) as response:
                response_data = await response.json()
                return response_data


    async def update_api_inbox(self, account_id: int, inbox_id: int, webhook_url: str) -> Dict[str, Any]:
        url = f"{self.base_url}/api/v1/accounts/{account_id}/inboxes/{inbox_id}"
        log.debug(f"Updating inbox at Chatwoot url: {url}")

        payload = {
            'enable_auto_assignment': False,
            'channel': {
                'webhook_url': webhook_url
            }
        }

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl)) as session:
            async with session.put(url, json=payload, headers=self.headers) as response:
                response_data = await response.json()
                return response_data


""" LIVEKIT Connector
    This connector is designed to work with LiveKit, a WebRTC platform for real-time communication.
    It provides a FastAPI router for handling chat completions and streaming responses.

    This connector will use DIRECT stream mode to send messages to the platform. Core events,
    and Middleware events are available.
"""


from fastapi.responses import StreamingResponse
from loguru import logger as log
from fastapi import APIRouter, BackgroundTasks, Request
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.message_gateway_context import MessageGatewayContext
from cel.connectors.livekit.model.livekit_message import LiveKitMessage
from cel.gateway.message_gateway import StreamMode
from cel.gateway.model.outgoing import OutgoingMessage

class LiveKitConnector(BaseConnector):
    """
    LiveKit connector that provides a FastAPI router for handling chat completions.
    """
    
    endpoint = '/chat/completions'
    
    def __init__(self):
        log.debug("Creating LiveKit connector")
        self.prefix = '/livekit'
        self.router = APIRouter(prefix=self.prefix)
        self.paused = False
        self.gateway = None
        self.__create_routes(self.router)
    
    def __create_routes(self, router: APIRouter):
        @router.post(f"{LiveKitConnector.endpoint}")
        async def chat_completion(request: Request, background_tasks: BackgroundTasks):
            data = await request.json()
            log.debug(f"[LiveKit] Incoming JSON: {data}")
            return StreamingResponse(self.__process_stream(data), media_type='text/event-stream')

    async def __process_stream(self, payload: dict):
        try:
            log.debug(f"[LiveKit] Processing stream with payload: {payload}")

            msg = await LiveKitMessage.load_from_message(payload, connector=self)
            
            if self.gateway:
            
                async for chunk in self.gateway.process_message(msg, mode=StreamMode.DIRECT, capture_response=True):
                    
                    # Send the content with proper formatting
                    if chunk.content:
                        yield f"data: {chunk.content}\n\n"
                    
                    # If this is the end of a partial chunk, add the flush sentinel
                    if not chunk.is_partial:
                        yield f"data: <FLUSH>\n\n"
            
            # End of stream
            yield f"data: \n\n"

        except Exception as e:
            log.error(f"Error processing LiveKit stream: {e}")
            yield f"data: Error: {str(e)}\n\n"

    async def send_typing_action(self, lead):
        # LiveKit doesn't support typing indicators
        log.debug("Typing indicators not supported in LiveKit")
        pass
    
    async def send_message(self, message: OutgoingMessage):
        # This method is not needed for the streaming interface
        log.debug("Direct message sending not used in LiveKit connector")
        pass
    
    def name(self) -> str:
        return "livekit"
    
    def get_router(self) -> APIRouter:
        return self.router
    
    def set_gateway(self, gateway):
        from cel.gateway.message_gateway import MessageGateway
        assert isinstance(gateway, MessageGateway), \
            "gateway must be an instance of MessageGateway"
        self.gateway = gateway

    def startup(self, context: MessageGatewayContext):
        # Verify if the webhook_url is set and is HTTPS
        if not context.webhook_url:
            log.warning("webhook_url is not set in the context - make sure your endpoint is accessible")
            return
            
        if not context.webhook_url.startswith("https"):
            log.warning(f"webhook_url should be HTTPS for production, got: {context.webhook_url}")
        
        webhook_url = f"{context.webhook_url}{self.prefix}{LiveKitConnector.endpoint}"
        log.info(f"Starting LiveKit connector with endpoint: {webhook_url}")
        log.info("Use this URL in the LiveKit Assistant settings under Model → Custom LLM URL")

    def shutdown(self, context: MessageGatewayContext):
        log.debug("Shutting down LiveKit connector")
    
    def pause(self):
        log.debug("Pausing LiveKit connector")
        self.paused = True

    def resume(self):
        log.debug("Resuming LiveKit connector")
        self.paused = False


from redis import Redis
from .base_cache import BaseCache

class RedisCache(BaseCache):
    def __init__(self, redis: str | Redis = None):
        self.client = redis if isinstance(redis, Redis) else Redis.from_url(redis or 'redis://localhost:6379/0')

    def memoize(self, typed: bool, tag: str, expire: int = None):
        def decorator(func):
            def wrapper(*args, **kwargs):
                key = f"{tag}:{args}:{kwargs}" if typed else f"{tag}:{args}"
                cached_result = self.client.get(key)
                # decode the cached result
                if cached_result:
                    cached_result = cached_result.decode('utf-8')
                    return cached_result
                
                result = func(*args, **kwargs)
                
                if result is not None and expire:
                    self.client.setex(key, expire, str(result))
                else:
                    self.client.set(key, str(result))
                    
                return result
                
            return wrapper
        return decorator


from abc import ABC
from dataclasses import dataclass
from fastapi import APIRouter, FastAPI


@dataclass
class MessageGatewayContext(ABC):
    router: APIRouter
    webhook_url: str
    app: FastAPI
    


import json
from langchain_openai import ChatOpenAI
from loguru import logger as log
from langsmith import traceable
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.outgoing import OutgoingMessage,\
                                            outgoing_message_from_dict,\
                                            OutgoingLinkMessage,\
                                            OutgoingSelectMessage,\
                                            OutgoingTextMessage



        
class SmartMessageEnhancerOpenAI:
    """ Enhances genAI output messages using OpenAI's GPT-3.5-turbo model.
    This will translate plain text messages into message structures 
    according to what the message wants to communicate. For example, 
    a message which is prompting the user to make a choice will be converted 
    into a message of type input_select with the information from the text. 
    The result must be a JSON.
    Later input_select will be converted by the gateway into a specific platform message 
    (e.g. a Telegram keyboard or a WhatsApp quick reply with the options).
    In order exploit the full potential of channels like WhatsApp and Telegram, 
    it is important to provide the right message structure.
    
    A message enhancer must be fail safe. If the enhancer fails to generate a message,
    it should return a simple text message with the original content.
    
    This enhancer is based on OpenAI's GPT-3.5-turbo model. If you want to use a different model,
    you can change the model parameter in the constructor.
    
    If you want to use a different fundational model, you can implement a new enhancer. 
    The enhancer should implement the __call__ method in this way:
    
    ```python
        async def __call__(self, lead: ConversationLead, 
                  text: str, 
                  is_partial: bool = True) -> OutgoingMessage:
                  
            ...your code here...
    ```
    
    """
    
    def __init__(self, model="gpt-3.5-turbo", only_full_messages=True, **kwargs):
        
        self.llm = ChatOpenAI(
            model=model,
            temperature=0,
            max_tokens=None,
            timeout=None,
            max_retries=2,
            streaming=False,
            **kwargs
        )
        self.only_full_messages = only_full_messages

        self.prompt = f"""You will convert plain text messages into message structures,
according to what the message wants to communicate.
For example, if the message asks the user to make a choice,
you will generate a message of type input_select with the information from the text.
The result must be a JSON.
                
Available message templates:
    {OutgoingTextMessage.description()}
    {OutgoingLinkMessage.description()}
    {OutgoingSelectMessage.description()}
"""
                
    async def __call__(self, lead: ConversationLead, 
                  text: str, 
                  is_partial: bool = True) -> OutgoingMessage:

        @traceable        
        async def smart_message_adapter():
            if is_partial and self.only_full_messages:
                log.debug(f"Skiping partial message: {text}")
                return OutgoingTextMessage(
                    lead=lead,
                    content=text
                )
            
            assert isinstance(lead, ConversationLead), \
                "lead must be an instance of ConversationLead"
            
            try:
                log.debug(f"Enhancing message: {text}")
                messages = [
                    (
                        "system",
                        f"{self.prompt}",
                    ),
                    ("system", f"Text to convert: {text}"),
                ]
                ai_msg = await self.llm.ainvoke(messages)
                log.debug(f"genAI enhanced message: {ai_msg}")
                # try to load json
                msg = json.loads(ai_msg.content)
                msg["lead"]=lead
                res = outgoing_message_from_dict(msg)
                return res

            except Exception as e:
                log.error(f"Error in enhance: {e}")
                
                # if there is an error, return a simple text message
                return OutgoingTextMessage(
                    lead=lead,
                    content=text
                )
                
        return await smart_message_adapter()





from cel.gateway.model.base_connector import BaseConnector
from cel.connectors.whatsapp.model.whatsapp_attachment import WhatsappAttachment
from cel.connectors.whatsapp.model.whatsapp_lead import WhatsappLead
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message


class WhatsappMessage(Message):
    
    def __init__(self, 
                 lead: ConversationLead, 
                 text: str = None, 
                 metadata: dict = None, 
                 date: int = None,
                 attachments: list[WhatsappAttachment] = None,
                 id: str = None
                ):
        super().__init__(lead, 
                         text=text, 
                         date=date, 
                         metadata=metadata, 
                         attachments=attachments, 
                         id=id)
    
    
    def is_voice_message(self):
        # TODO: Implement voice message check
        return False
  
    
    
    @classmethod
    async def load_from_message(cls, 
                                data: dict, 
                                connector: BaseConnector, 
                                token: str,
                                phone_number_id: str = None):
        
        assert token is not None, "token must be provided for attachment download"
        
        assert isinstance(data, dict), "data must be a dictionary"
        assert isinstance(connector, BaseConnector), "connector must be an instance of BaseConnector"
        assert "entry" in data, "data must have an entry key"
        
        entry = data.get("entry")
        assert isinstance(entry, list), "entry must be a list"
        assert len(entry) > 0, "entry must have at least one element"
        
        entry0 = entry[0]
        assert isinstance(entry0, dict), "first entry must be a dictionary"
        assert "changes" in entry0, "first entry must have a changes key"
        
        changes = entry0.get("changes")
        assert isinstance(changes, list), "changes must be a list"
        assert len(changes) > 0, "changes must have at least one element"
        
        change0 = changes[0]
        assert isinstance(change0, dict), "first change must be a dictionary"
        
        change0 = change0.get("value")
        msg0 = change0.get("messages")[0]

        # get text from message or caption if it is a media message
        text = msg0.get("text", {}).get("body") 
        
        type = msg0.get("type")
        if type == 'interactive':
            interactive = msg0.get("interactive", {})
            if interactive.get("type") == 'button_reply':
                text = interactive.get("button_reply", {}).get("title")
            # {'type': 'list_reply', 'list_reply': {'id': '2', 'title': '2'}}
            if interactive.get("type") == 'list_reply':
                text = interactive.get("list_reply", {}).get("title")
        
        date = int(msg0.get("timestamp"))
        id = msg0.get("id")
        metadata = {}
        lead = WhatsappLead.from_whatsapp_message(data, connector=connector)
        attach = await WhatsappAttachment.load_from_message(data, token, phone_number_id)
        return WhatsappMessage(lead=lead,
                               id=id,
                               text=text,
                               date=date,
                               metadata=metadata,
                               attachments=[attach] if attach else None
                            )
        

    def __str__(self):
        return f"TelegramMessage: {self.text}"
        
    def __repr__(self):
        return f"TelegramMessage: {self.text}"
    
    


from typing import Any, Dict, Union

@staticmethod
def is_reaction(data: Dict[Any, Any]) -> bool:
    """is_reaction checks if the data received from the webhook is a message.

    Args:
        data (Dict[Any, Any]): The data received from the webhook

    Returns:
        bool: True if the data is a message, False otherwise
    """

    msg = data.get("entry", [{}])[0].get("changes", [{}])[0].get("value", {}).get("messages", [{}])[0]
    if "reaction" in msg:
        return True
    else:
        return False

@staticmethod
def is_message(data: Dict[Any, Any]) -> bool:
    """is_message checks if the data received from the webhook is a message.

    Args:
        data (Dict[Any, Any]): The data received from the webhook

    Returns:
        bool: True if the data is a message, False otherwise
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "messages" in data:
        return True
    else:
        return False


@staticmethod
def get_mobile(data: Dict[Any, Any]) -> Union[str, None]:
    """
    Extracts the mobile number of the sender from the data received from the webhook.

    Args:
        data[dict]: The data received from the webhook
    Returns:
        str: The mobile number of the sender

    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> mobile = whatsapp.get_mobile(data)
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "contacts" in data:
        return data["contacts"][0]["wa_id"]


@staticmethod
def get_name(data: Dict[Any, Any]) -> Union[str, None]:
    """
    Extracts the name of the sender from the data received from the webhook.

    Args:
        data[dict]: The data received from the webhook
    Returns:
        str: The name of the sender
    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> mobile = whatsapp.get_name(data)
    """
    contact = data["entry"][0]["changes"][0]["value"]
    if contact:
        return contact["contacts"][0]["profile"]["name"]


@staticmethod
def get_message(data: Dict[Any, Any]) -> Union[str, None]:
    """
    Extracts the text message of the sender from the data received from the webhook.

    Args:
        data[dict]: The data received from the webhook
    Returns:
        str: The text message received from the sender
    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> message = message.get_message(data)
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "messages" in data:
        return data["messages"][0]["text"]["body"]


@staticmethod
def get_message_id(data: Dict[Any, Any]) -> Union[str, None]:
    """
    Extracts the message id of the sender from the data received from the webhook.

    Args:
        data[dict]: The data received from the webhook
    Returns:
        str: The message id of the sender
    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> message_id = whatsapp.get_message_id(data)
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "messages" in data:
        return data["messages"][0]["id"]


@staticmethod
def get_message_timestamp(data: Dict[Any, Any]) -> Union[str, None]:
    """ "
    Extracts the timestamp of the message from the data received from the webhook.

    Args:
        data[dict]: The data received from the webhook
    Returns:
        str: The timestamp of the message
    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> whatsapp.get_message_timestamp(data)
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "messages" in data:
        return data["messages"][0]["timestamp"]


@staticmethod
def get_interactive_response(data: Dict[Any, Any]) -> Union[Dict, None]:
    """
        Extracts the response of the interactive message from the data received from the webhook.

        Args:
        data[dict]: The data received from the webhook
    Returns:
        dict: The response of the interactive message

    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> response = whatsapp.get_interactive_response(data)
        >>> interactive_type = response.get("type")
        >>> message_id = response[interactive_type]["id"]
        >>> message_text = response[interactive_type]["title"]
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "messages" in data:
        if "interactive" in data["messages"][0]:
            return data["messages"][0]["interactive"]


@staticmethod
def get_location(data: Dict[Any, Any]) -> Union[Dict, None]:
    """
    Extracts the location of the sender from the data received from the webhook.

    Args:
        data[dict]: The data received from the webhook

    Returns:
        dict: The location of the sender

    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> whatsapp.get_location(data)
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "messages" in data:
        if "location" in data["messages"][0]:
            return data["messages"][0]["location"]


@staticmethod
def get_image(data: Dict[Any, Any]) -> Union[Dict, None]:
    """ "
    Extracts the image of the sender from the data received from the webhook.

    Args:
        data[dict]: The data received from the webhook
    Returns:
        dict: The image_id of an image sent by the sender

    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> image_id = whatsapp.get_image(data)
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "messages" in data:
        if "image" in data["messages"][0]:
            return data["messages"][0]["image"]


@staticmethod
def get_document(data: Dict[Any, Any]) -> Union[Dict, None]:
    """ "
    Extracts the document of the sender from the data received from the webhook.

    Args:
        data[dict]: The data received from the webhook
    Returns:
        dict: The document_id of an image sent by the sender

    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> document_id = whatsapp.get_document(data)
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "messages" in data:
        if "document" in data["messages"][0]:
            return data["messages"][0]["document"]


@staticmethod
def get_audio(data: Dict[Any, Any]) -> Union[Dict, None]:
    """
    Extracts the audio of the sender from the data received from the webhook.

    Args:
        data[dict]: The data received from the webhook

    Returns:
        dict: The audio of the sender

    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> whatsapp.get_audio(data)
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "messages" in data:
        if "audio" in data["messages"][0]:
            return data["messages"][0]["audio"]


@staticmethod
def get_video(data: Dict[Any, Any]) -> Union[Dict, None]:
    """
    Extracts the video of the sender from the data received from the webhook.

    Args:
        data[dict]: The data received from the webhook

    Returns:
        dict: Dictionary containing the video details sent by the sender

    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> whatsapp.get_video(data)
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "messages" in data:
        if "video" in data["messages"][0]:
            return data["messages"][0]["video"]


@staticmethod
def get_message_type(data: Dict[Any, Any]) -> Union[str, None]:
    """
    Gets the type of the message sent by the sender from the data received from the webhook.


    Args:
        data [dict]: The data received from the webhook

    Returns:
        str: The type of the message sent by the sender

    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> whatsapp.get_message_type(data)
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "messages" in data:
        return data["messages"][0]["type"]


@staticmethod
def get_delivery(data: Dict[Any, Any]) -> Union[Dict, None]:
    """
    Extracts the delivery status of the message from the data received from the webhook.
    Args:
        data [dict]: The data received from the webhook

    Returns:
        dict: The delivery status of the message and message id of the message
    """
    data = data["entry"][0]["changes"][0]["value"]
    if "statuses" in data:
        return data["statuses"][0]["status"]


@staticmethod
def changed_field(data: Dict[Any, Any]) -> str:
    """
    Helper function to check if the field changed in the data received from the webhook.

    Args:
        data [dict]: The data received from the webhook

    Returns:
        str: The field changed in the data received from the webhook

    Example:
        >>> from whatsapp import WhatsApp
        >>> whatsapp = WhatsApp(token, phone_number_id)
        >>> whatsapp.changed_field(data)
    """
    return data["entry"][0]["changes"][0]["field"]


@staticmethod
def get_author(data: Dict[Any, Any]) -> Union[str, None]:
    try:
        return data["entry"][0]["changes"][0]["value"]["messages"][0]["from"]
    except Exception:
        return None


